package {

import flash.display.Sprite;
//import flash.display.Shape;
import flash.text.TextFormat;
import flash.text.TextField;
import flash.text.TextFieldAutoSize;


import Obstacle;
import Log;

// Class for an individual track instance.
public class Track extends Sprite {

  static public var WINDOW_WIDTH:Number = 800;
  static public var WINDOW_HEIGHT:Number = 600;
  static public var LINE_THICKNESS:Number = 10;
  static public var LOST_POINTS_ON_FINISH:Number = 1;

  static private var LINE_HEIGHT:Number = WINDOW_HEIGHT - (WINDOW_HEIGHT / 3);

  static private var SCROLL_SPEED:Number = 5;

  static private var RUNNER_SPEED:Number = 8;
  static private var INITIAL_RUNNER_DISTANCE:Number = 200;
  static private var MAX_RUNNER_LEAD_DISTANCE:Number = 600;
  static private var RUNNER_WIDTH:Number = 30;
  static private var RUNNER_HEIGHT:Number = 70;
  static private var RUNNER_JUMP_INTERVAL:Number = 15;
  static private var RUNNER_APEX_INTERVAL:Number = 7;
  static private var RUNNER_JUMP_HEIGHT:Number = 50;
  static private var RUNNER_JUMP_SPEED:Number = 30;
  static private var RUNNER_OBSTACLE_GAP:Number = 60;
  static private var RUNNER_SKILL_LOW:Number = 1;
  static private var RUNNER_SKILL_MEDIUM:Number = 2;
  static private var RUNNER_SKILL_HIGH:Number = 3;
  static private var JUMP_CLEAR_DISTANCE:Number = 5;

  static private var INITIAL_OBSTACLE_DISTANCE:Number = 810;
  static private var MIN_OBSTACLE_DISTANCE:Number = 200;
  static private var MAX_OBSTACLE_DISTANCE:Number = 500;
  static private var NUM_ACTIVE_OBSTACLES:Number = 5;

  static private var SCORE_DISPLAY_STARTING_X:Number = 30;
  static private var SCORE_DISPLAY_STARTING_Y:Number = 300;
  static private var SCORE_DISPLAY_FONT_SIZE:Number = 130;
  static private var SCORE_DISPLAY_RISE_SPEED:Number = 10;
  static private var SCORE_DISPLAY_ALPHA_START:Number = 15;
  static private var SCORE_DISPLAY_ALPHA_STEP:Number = 0.08;

  static private var SMALL_OBSTACLE_POINTS:Number = 1;
  static private var MEDIUM_OBSTACLE_POINTS:Number = 4;
  static private var LARGE_OBSTACLE_POINTS:Number = 9;

  static private var LOST_POINTS_COLOR:Number = 0xffff00;
  static private var LOST_SCORE_DISPLAY_STARTING_X:Number = 330;
  static private var LOST_SCORE_DISPLAY_STARTING_Y:Number = 200;

  static private var WARNING_DISTANCE:Number = 250;

  //  static private var inactiveObstacles:Vector.<Obstacle>;

  private var scrollDistance:Number;
  private var maxObstacleDistance:Number;
  private var runnerDistance:Number;
  private var currentRunnerObstacle:Number;
  //  private var trackLine:Shape;
  private var activeObstacles:Vector.<Obstacle>;
  private var trackFinished:Boolean;
  private var isJumping:Boolean;
  private var jumpCounter:Number;
  private var apexCounter:Number;
  private var targetHeight:Number;
  private var runnerHeight:Number;
  private var runnerVerticalDirection:Number;
  private var runnerSkill:Number;
  private var obstaclesCleared:Number;
  private var trackScore:Number;
  private var scoreDisplay:TextField;
  private var currentlyDisplayingScore:Boolean;
  private var scoreDisplayTick:Number;
  private var displayingLostPoints:Boolean;

  // Application entry point.
  public function Track() {
    //    scrollDistance = 0;
    //    runnerDistance = INITIAL_RUNNER_DISTANCE;
    //    currentRunnerObstacle = -1;
    //    trackFinished = false;

    var format:TextFormat = new TextFormat();
    format.font = "HurdlesFont";
    format.size = SCORE_DISPLAY_FONT_SIZE;
    format.bold = true;
    //    format.align = TextFormatAlign.CENTER;

    scoreDisplay = new TextField();
    scoreDisplay.autoSize = TextFieldAutoSize.LEFT;
    scoreDisplay.defaultTextFormat = format;
    scoreDisplay.selectable = false;
    scoreDisplay.embedFonts = true;
    addChild(scoreDisplay);

    displayingLostPoints = false;

    activeObstacles = new Vector.<Obstacle>;
    reset();

    //    maxObstacleDistance = INITIAL_OBSTACLE_DISTANCE;
    //    activeObstacles.push(new Obstacle(maxObstacleDistance));

    //    updateObstacles();
  }

  public function reset():void {
    scrollDistance = 0;
    runnerDistance = INITIAL_RUNNER_DISTANCE;
    currentRunnerObstacle = -1;
    trackFinished = false;
    activeObstacles.length = 0;
    maxObstacleDistance = INITIAL_OBSTACLE_DISTANCE;
    isJumping = false;
    jumpCounter = 0;
    runnerHeight = 0;
    runnerVerticalDirection = 1;
    runnerSkill = RUNNER_SKILL_LOW;
    obstaclesCleared = 0;
    apexCounter = 0;
    targetHeight = 0;
    trackScore = 0;
    currentlyDisplayingScore = false;
    scoreDisplayTick = 0;
    scoreDisplay.x = SCORE_DISPLAY_STARTING_X;
    scoreDisplay.y = SCORE_DISPLAY_STARTING_Y;
    scoreDisplay.textColor = 0x000000;
    scoreDisplay.alpha = 1.0;
    scoreDisplay.visible = false;
    scoreDisplay.text = "";
    activeObstacles.push(new Obstacle(maxObstacleDistance));
    //    updateObstacles();
  }

  public function update():void {

    //    var halfLineThickness:Number = LINE_THICKNESS / 2;

    //    graphics.moveTo(halfLineThickness, halfLineThickness);
    //    graphics.lineTo(WINDOW_WIDTH - halfLineThickness, halfLineThickness);
    //    graphics.lineTo(WINDOW_WIDTH - halfLineThickness, WINDOW_HEIGHT - halfLineThickness);
    //    graphics.lineTo(halfLineThickness, WINDOW_HEIGHT - halfLineThickness);
    //    graphics.lineTo(halfLineThickness, halfLineThickness);

    //    if(trackFinished) {
      //      graphics.clear();
      //      graphics.beginFill(0x111111, 1);
      //      graphics.drawRect(0, 0, WINDOW_WIDTH, WINDOW_HEIGHT);
      //      graphics.endFill();
      //    }
    //    else {


    if(displayingLostPoints) {
      ++scoreDisplayTick;
      scoreDisplay.y += SCORE_DISPLAY_RISE_SPEED;

      if(scoreDisplayTick >= SCORE_DISPLAY_ALPHA_START) {
	scoreDisplay.alpha = Math.max(scoreDisplay.alpha - SCORE_DISPLAY_ALPHA_STEP, 0);

	if(scoreDisplay.alpha == 0) {
	  displayingLostPoints = false;
	  scoreDisplay.visible = false;
	  //	  scoreDisplay.alpha = 1.0;
	}
      }
    }

    if(trackFinished) {
      return;
    }

    scrollDistance += SCROLL_SPEED;

    var oldRunnerDistance:Number = runnerDistance;
    runnerDistance += RUNNER_SPEED;

    updateObstacles();

    graphics.clear();
    graphics.lineStyle(LINE_THICKNESS, 0x000000);

    if((runnerDistance - scrollDistance) < WARNING_DISTANCE && currentRunnerObstacle != -1) {
      graphics.beginFill(0xffff00, 1);
    }
    else {
      graphics.beginFill(0xffffff, 1);
    }

    graphics.drawRect(0, 0, WINDOW_WIDTH, WINDOW_HEIGHT);
    graphics.endFill();

    //    graphics.moveTo(0, 0);
    //    graphics.lineTo(WINDOW_WIDTH, 0);
    //    graphics.lineTo(WINDOW_WIDTH, WINDOW_HEIGHT);
    //    graphics.lineTo(0, WINDOW_HEIGHT);
    //    graphics.lineTo(0, 0);

    graphics.moveTo(0, LINE_HEIGHT);

    var drawFinalLine:Boolean = true;
    var obstacleColor:Number = 0x000000;

    for(var i:int = 0; i < activeObstacles.length; ++i) {
      var obstaclePosition:Number = activeObstacles[i].getLinePosition();
      var obstacleLeftBorder:Number = obstaclePosition - scrollDistance;

      if(obstacleLeftBorder > WINDOW_WIDTH) {
	break;
      }

      var obstacleHeight:Number = activeObstacles[i].getHeight();
      var obstacleUnitHeight:Number = activeObstacles[i].getUnitHeight();

      var obstacleTopBorder:Number = LINE_HEIGHT - obstacleHeight;
      var obstacleRightBorder:Number = obstacleLeftBorder + Obstacle.OBSTACLE_WIDTH;

      obstacleColor = 0x000000;

      if(obstacleUnitHeight == Obstacle.HEIGHT_SMALL) {
	obstacleColor = Obstacle.OBSTACLE_COLOR_SMALL;
      }
      else if(obstacleUnitHeight == Obstacle.HEIGHT_MEDIUM) {
	obstacleColor = Obstacle.OBSTACLE_COLOR_MEDIUM;
      }
      else if(obstacleUnitHeight == Obstacle.HEIGHT_LARGE) {
	obstacleColor = Obstacle.OBSTACLE_COLOR_LARGE;
      }

      graphics.lineStyle(LINE_THICKNESS, 0x000000);

      if(obstacleLeftBorder < 0) {
	graphics.beginFill(obstacleColor, 1);
	graphics.drawRect(0, obstacleTopBorder, Obstacle.OBSTACLE_WIDTH + obstacleLeftBorder, obstacleHeight);
	graphics.endFill();
	graphics.moveTo(obstacleRightBorder, LINE_HEIGHT);
      }
      else if(obstacleRightBorder > WINDOW_WIDTH) {
	graphics.lineTo(obstacleLeftBorder, LINE_HEIGHT);
	graphics.beginFill(obstacleColor, 1);
	graphics.drawRect(obstacleLeftBorder, obstacleTopBorder, WINDOW_WIDTH - obstacleLeftBorder, obstacleHeight);
	graphics.endFill();
	graphics.moveTo(WINDOW_WIDTH, obstacleTopBorder);
	drawFinalLine = false;
      }
      else {
	graphics.lineTo(obstacleLeftBorder, LINE_HEIGHT);
	graphics.beginFill(obstacleColor, 1);
	graphics.drawRect(obstacleLeftBorder, obstacleTopBorder, Obstacle.OBSTACLE_WIDTH, obstacleHeight);
	graphics.endFill();
	graphics.moveTo(obstacleRightBorder, LINE_HEIGHT);
      }

      var runnerStoppingPoint:Number = obstaclePosition - RUNNER_OBSTACLE_GAP;

      if(oldRunnerDistance <= runnerStoppingPoint && runnerDistance > runnerStoppingPoint) {
	currentRunnerObstacle = i;
      }

      var obstaclePassed:Boolean = false;

      // Handle runner
      if(i == currentRunnerObstacle) {
	//	if(isJumping) {
	//	  runnerHeight += runnerVerticalDirection * RUNNER_JUMP_SPEED;
	//
	//	  if(runnerVerticalDirection > 0) {
	//	    if(runnerHeight >= RUNNER_JUMP_HEIGHT) {
	//	      runnerHeight = RUNNER_JUMP_HEIGHT;
	//	      runnerVerticalDirection *= -1;
	//	    }
	//	  }
	//	  else {
	//	    if(runnerHeight <= 0) {
	//	      runnerHeight = 0;
	//	      runnerVerticalDirection *= -1;
	//	      isJumping = false;
	//	    }
	//	  }
	//
	//	  //	  runnerHeight = 0;
	//	}

	if(!isJumping) {
	  ++jumpCounter;

	  if(jumpCounter == RUNNER_JUMP_INTERVAL) {
	    // TODO: This is where we need to roll the dice to see if we cleared the obstacle
	    if(runnerSkill >= obstacleUnitHeight) {
	      // Flip a coin for now
	      if(Math.floor(Math.random() * 2) == 1) {
		currentRunnerObstacle = -1;
		obstaclePassed = true;
		targetHeight = obstacleHeight + RUNNER_HEIGHT + JUMP_CLEAR_DISTANCE;
		++obstaclesCleared;

		if(runnerSkill == RUNNER_SKILL_LOW && obstaclesCleared >= 3) {
		  runnerSkill = RUNNER_SKILL_MEDIUM;
		}
		else if(runnerSkill == RUNNER_SKILL_MEDIUM && obstaclesCleared >= 8) {
		  runnerSkill = RUNNER_SKILL_HIGH;
		}

	      }
	      else {
		targetHeight = RUNNER_JUMP_HEIGHT;
	      }
	    }
	    else {
	      targetHeight = RUNNER_JUMP_HEIGHT;
	    }

	    isJumping = true;
	    jumpCounter = 0;
	    apexCounter = 0;
	  }
	}

	if(!obstaclePassed) {
	  runnerDistance = runnerStoppingPoint;
	}
      }

      //      if(obstacleLeftBorder >= 0) {
      //	//	graphics.lineTo(obstacleLeftBorder, LINE_HEIGHT);
      //	//	graphics.lineTo(obstacleLeftBorder, obstacleTopBorder);
      //
      //	graphics.beginFill(obstacleColor, 1);
      //	graphics.drawRect(obstacleLeftBorder, obstacleTopBorder, Obstacle.OBSTACLE_WIDTH, obstacleHeight);
      //	graphics.endFill();
      //      }
      //      else {
      //	graphics.beginFill(obstacleColor, 1);
      //	graphics.drawRect(0, obstacleTopBorder, Obstacle.OBSTACLE_WIDTH + obstacleLeftBorder, obstacleHeight);
      //	graphics.endFill();
      //      }

      //      graphics.lineStyle(LINE_THICKNESS, 0x000000);
      //      graphics.moveTo(obstacleLeftBorder, obstacleTopBorder);
      //
      //      if(obstacleRightBorder > WINDOW_WIDTH) {
      //	graphics.lineTo(WINDOW_WIDTH, obstacleTopBorder);
      //	drawFinalLine = false;
      //	break;
      //      }
      //      else {
      //	graphics.lineTo(obstacleRightBorder, obstacleTopBorder);
      //	graphics.lineTo(obstacleRightBorder, LINE_HEIGHT);
      //      }
    }

    if(drawFinalLine) {
      graphics.lineStyle(LINE_THICKNESS, 0x000000);
      graphics.lineTo(WINDOW_WIDTH, LINE_HEIGHT);
    }

    // Cap runner to stay onscreen
    if(runnerDistance - scrollDistance > MAX_RUNNER_LEAD_DISTANCE) {
      runnerDistance = scrollDistance + MAX_RUNNER_LEAD_DISTANCE;
    }

    // Determine runner y position
    if(isJumping) {
      runnerHeight += runnerVerticalDirection * RUNNER_JUMP_SPEED;

      if(runnerVerticalDirection > 0) {
	if(runnerHeight >= targetHeight) {
	  runnerHeight = targetHeight;

	  if(targetHeight == RUNNER_JUMP_HEIGHT) {
	    runnerVerticalDirection *= -1;
	  }
	  else {
	    ++apexCounter;

	    if(apexCounter >= RUNNER_APEX_INTERVAL) {
	      runnerVerticalDirection *= -1;
	    }
	  }


	  //	  runnerVerticalDirection *= -1;
	}
      }
      else {
	if(runnerHeight <= 0) {
	  runnerHeight = 0;
	  runnerVerticalDirection *= -1;
	  isJumping = false;
	}
      }

      //	  runnerHeight = 0;
    }
    //    else {
    //      ++jumpCounter;
    //      if(jumpCounter == RUNNER_JUMP_INTERVAL) {
    //	isJumping = true;
    //	jumpCounter = 0;
    //      }
    //    }

    // Draw runner
    if(runnerSkill == RUNNER_SKILL_LOW) {
      obstacleColor = Obstacle.OBSTACLE_COLOR_SMALL;
    }
    else if(runnerSkill == RUNNER_SKILL_MEDIUM) {
      obstacleColor = Obstacle.OBSTACLE_COLOR_MEDIUM;
    }
    else if(runnerSkill == RUNNER_SKILL_HIGH) {
      obstacleColor = Obstacle.OBSTACLE_COLOR_LARGE;
    }

    graphics.lineStyle(LINE_THICKNESS, 0x000000);
    graphics.beginFill(obstacleColor, 1);
    graphics.drawEllipse(runnerDistance - scrollDistance, LINE_HEIGHT - RUNNER_HEIGHT - runnerHeight, RUNNER_WIDTH, RUNNER_HEIGHT);
    graphics.endFill();

    if(runnerDistance < scrollDistance) {
      trackFinished = true;
      currentlyDisplayingScore = false;
      scoreDisplay.visible = false;
      graphics.clear();
      graphics.lineStyle(LINE_THICKNESS, 0x111111);
      graphics.beginFill(0x111111, 1);
      graphics.drawRect(0, 0, WINDOW_WIDTH, WINDOW_HEIGHT);
      graphics.endFill();

      displayingLostPoints = true;
      scoreDisplay.textColor = LOST_POINTS_COLOR;
      scoreDisplay.text = "-" + String(LOST_POINTS_ON_FINISH);
      scoreDisplay.x = LOST_SCORE_DISPLAY_STARTING_X;
      scoreDisplay.y = LOST_SCORE_DISPLAY_STARTING_Y;
      scoreDisplay.alpha = 1.0;
      scoreDisplay.visible = true;
    }

    if(currentlyDisplayingScore) {
      ++scoreDisplayTick;
      scoreDisplay.y -= SCORE_DISPLAY_RISE_SPEED;

      if(scoreDisplayTick >= SCORE_DISPLAY_ALPHA_START) {
	scoreDisplay.alpha = Math.max(scoreDisplay.alpha - SCORE_DISPLAY_ALPHA_STEP, 0);

	if(scoreDisplay.alpha == 0) {
	  currentlyDisplayingScore = false;
	  scoreDisplay.visible = false;
	  //	  scoreDisplay.alpha = 1.0;
	}
      }
    }

    //  static private var SCORE_DISPLAY_STARTING_X:Number = 30;
    //  static private var SCORE_DISPLAY_STARTING_Y:Number = 300;
    //  static private var SCORE_DISPLAY_FONT_SIZE:Number = 130;
    //  static private var SCORE_DISPLAY_RISE_SPEED:Number = 4;
    //  static private var SCORE_DISPLAY_ALPHA_START:Number = 10;
    //  static private var SCORE_DISPLAY_ALPHA_STEP:Number = 10;

    //    }


  }

  private function updateObstacles():void {
    //    Log.print("Updating obstacles");

    // Remove old obstacles
    for(var j:int = 0; j < activeObstacles.length; ) {
      var currentObstacleUnitHeight:Number = activeObstacles[j].getUnitHeight();

      if((activeObstacles[j].getLinePosition() + Obstacle.OBSTACLE_WIDTH < scrollDistance) || currentObstacleUnitHeight == 0) {

	// Score points
	if(currentObstacleUnitHeight > 0) {
	  currentlyDisplayingScore = true;
	  scoreDisplayTick = 0;

	  if(currentObstacleUnitHeight == Obstacle.HEIGHT_SMALL) {
	    trackScore += SMALL_OBSTACLE_POINTS;
	    scoreDisplay.textColor = Obstacle.OBSTACLE_COLOR_SMALL;
	    scoreDisplay.text = "+" + String(SMALL_OBSTACLE_POINTS);
	  }
	  else if(currentObstacleUnitHeight == Obstacle.HEIGHT_MEDIUM) {
	    trackScore += MEDIUM_OBSTACLE_POINTS;
	    scoreDisplay.textColor = Obstacle.OBSTACLE_COLOR_MEDIUM;
	    scoreDisplay.text = "+" + String(MEDIUM_OBSTACLE_POINTS);
	  }
	  else if(currentObstacleUnitHeight == Obstacle.HEIGHT_LARGE) {
	    trackScore += LARGE_OBSTACLE_POINTS;
	    scoreDisplay.textColor = Obstacle.OBSTACLE_COLOR_LARGE;
	    scoreDisplay.text = "+" + String(LARGE_OBSTACLE_POINTS);
	  }

	  scoreDisplay.x = SCORE_DISPLAY_STARTING_X;
	  scoreDisplay.y = SCORE_DISPLAY_STARTING_Y;
	  scoreDisplay.alpha = 1.0;
	  scoreDisplay.visible = true;
	}

	activeObstacles.splice(j, 1);
	//	Log.print("Obstacle removed");

	if(currentRunnerObstacle != -1) {
	  --currentRunnerObstacle;
	}

	//	if(j == currentRunnerObstacle) {
	//	  currentRunnerObstacle = -1;
	//	}
      }
      else {
	++j;
	//	break;
      }
    }

    // Add new obstacles
    for(var k:int = activeObstacles.length; k < NUM_ACTIVE_OBSTACLES; ++k) {
      maxObstacleDistance += MIN_OBSTACLE_DISTANCE + Math.floor(Math.random() * (MAX_OBSTACLE_DISTANCE - MIN_OBSTACLE_DISTANCE));
      activeObstacles.push(new Obstacle(maxObstacleDistance));
      //      Log.print("Obstacle added");
    }
  }

  public function isFinished():Boolean {
    return trackFinished;
  }

  public function handleClick():void {
    // TODO: CHANGE THIS TO BE NEXT OBSTACLE IN FRONT OF RUNNER ONCE RUNNER CAN JUMP HURDLES!
    var nextObstacleIndex:Number = currentRunnerObstacle;

    if(nextObstacleIndex == -1) {
      for(var m:int = 0; m < activeObstacles.length; ++m) {
	if(activeObstacles[m].getLinePosition() - RUNNER_OBSTACLE_GAP > runnerDistance) {
	  nextObstacleIndex = m;
	  break;
	}
      }
    }

    activeObstacles[nextObstacleIndex].handleClick();

    if(activeObstacles[nextObstacleIndex].getUnitHeight() == 0) {
      currentRunnerObstacle = -1;
    }

    updateObstacles();
    //    if(activeObstacles[0].getUnitHeight() == 0) {
    //      activeObstacles.splice(0, 1);
    //    }
  }

  public function getScore():Number {
    return trackScore;
  }

}

}